# Copyright (c) 2025 Centre National d'Etudes Spatiales (CNES).
#
# This file is part of the xDEM project:
# https://github.com/glaciohack/xdem
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import argparse
import ctypes.util
import logging

import yaml  # type: ignore

from xdem.workflows import Accuracy, Topo
from xdem.workflows.schemas import COMPLETE_CONFIG_ACCURACY, COMPLETE_CONFIG_TOPO

lib_gobject_name = ctypes.util.find_library("gobject-2.0")
lib_pango_name = ctypes.util.find_library("pango-1.0")

if lib_gobject_name and lib_pango_name:
    from weasyprint import HTML

    _has_libgobject = True
else:
    _has_libgobject = False


def main() -> None:
    """
    Main function for the CLI
    """

    parser = argparse.ArgumentParser(prog="xdem", description="CLI tool to process DEM workflows")
    parser.add_argument(
        "--log-level",
        default="INFO",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        help="Set the logging level",
    )
    subparsers = parser.add_subparsers(dest="command", help="Available workflows as subcommand")

    # Subcommand: info
    topo_parser = subparsers.add_parser(
        "topo",
        help="Run DEM qualification workflow",
        description="Run a DEM information workflow using a YAML configuration file.",
        epilog="Example: xdem topo config.yaml",
    )
    topo_group = topo_parser.add_mutually_exclusive_group(required=True)
    topo_group.add_argument(
        "--config",
        help="Path to YAML configuration file",
    )
    topo_group.add_argument("--display_template_config", action="store_true", help="Show configuration template")

    # Subcommand: accuracy
    diff_parser = subparsers.add_parser(
        "accuracy",
        help="Run DEM comparison workflow",
        description="Run a DEM comparison workflow using a YAML configuration file.",
        epilog="Example: xdem accuracy config.yaml",
    )
    diff_group = diff_parser.add_mutually_exclusive_group(required=True)
    diff_group.add_argument("--config", help="Path to YAML configuration file")
    diff_group.add_argument("--display_template_config", action="store_true", help="Show configuration template")

    args = parser.parse_args()

    # Instance logger
    log_level = getattr(logging, args.log_level.upper(), logging.INFO)
    logging.basicConfig(level=log_level, format="%(asctime)s - %(levelname)s - %(message)s")
    logger = logging.getLogger(__name__)
    # fontTools creates noisy logs
    logging.getLogger("fontTools").setLevel(logging.WARNING)
    logging.getLogger("fontTools").propagate = False

    if args.command == "topo":
        if args.display_template_config:
            yaml_string = yaml.dump(COMPLETE_CONFIG_TOPO, sort_keys=False, allow_unicode=True)
            logging.info("\n" + yaml_string)
        elif args.config:
            logger.info("Running topo workflow")
            workflow = Topo(args.config)
            workflow.run()

    elif args.command == "accuracy":
        if args.display_template_config:
            yaml_string = yaml.dump(COMPLETE_CONFIG_ACCURACY, sort_keys=False, allow_unicode=True)
            logging.info("\n" + yaml_string)
        elif args.config:
            logger.info("Running accuracy workflow")
            workflow = Accuracy(args.config)  # type: ignore
            workflow.run()

    else:
        raise ValueError(f"{args.command} doesn't exist, valid command are 'accuracy', 'topo'")

    if args.config and _has_libgobject:
        logger.info("Generating HTML and PDF report")
        HTML(workflow.outputs_folder / "report.html").write_pdf(workflow.outputs_folder / "report.pdf")

    logger.info("End of execution")


if __name__ == "__main__":
    main()



# Copyright (c) 2025 Centre National d'Etudes Spatiales (CNES).
#
# This file is part of the xDEM project:
# https://github.com/glaciohack/xdem
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""

test for workflow class
"""
# mypy: disable-error-code=no-untyped-def
import csv

import numpy as np
import pytest
import yaml  # type: ignore

import xdem
from xdem.workflows.topo import Topo
from xdem.workflows.workflows import Workflows

pytestmark = pytest.mark.filterwarnings("ignore::UserWarning")


def test_workflows_init_wrong_config():
    """
    Test workflows class initialization
    """
    user_config = 2
    with pytest.raises(ValueError, match="The configuration should be provided either as a path"):
        _ = Topo(user_config)  # type: ignore


def test_workflows_init(pipeline_topo, get_topo_inputs_config, tmp_path):
    """
    Test workflows class initialization
    """
    user_config = get_topo_inputs_config
    user_config["outputs"] = {"path": str(tmp_path)}
    workflows = Topo(user_config)

    assert isinstance(workflows, Workflows)
    pipeline_gt = pipeline_topo
    pipeline_gt["outputs"] = {"path": str(tmp_path), "level": 1}
    assert workflows.config == pipeline_topo
    assert workflows.level == 1
    assert workflows.outputs_folder == tmp_path
    assert workflows.outputs_folder.exists()
    for folder in ["plots", "rasters", "tables"]:
        assert workflows.outputs_folder.joinpath(folder).exists()
    assert workflows.outputs_folder.joinpath("used_config.yaml").exists()
    assert workflows.dico_to_show == [
        (
            "Information about inputs",
            {
                "reference_elev": {
                    "path_to_elev": xdem.examples.get_path("longyearbyen_tba_dem"),
                    "path_to_mask": xdem.examples.get_path("longyearbyen_glacier_outlines"),
                    "from_vcrs": None,
                    "to_vcrs": None,
                }
            },
        )
    ]


@pytest.mark.parametrize(
    "input_data,expected",
    [
        ({"a": 1, "b": None, "c": 3}, {"a": 1, "c": 3}),
        ({"a": {"x": None, "y": 2}, "b": None}, {"a": {"y": 2}}),
        ([1, None, 2, None, 3], [1, 2, 3]),
        ([{"a": 1, "b": None}, {"c": None}, None, {"d": 4}], [{"a": 1}, {}, {"d": 4}]),
        (
            {"a": [1, None, {"b": None, "c": 3}], "d": None, "e": {"f": None, "g": [None, 7]}},
            {"a": [1, {"c": 3}], "e": {"g": [7]}},
        ),
        ({}, {}),
        ([], []),
        ({"a": 1, "b": [2, 3], "c": {"d": 4}}, {"a": 1, "b": [2, 3], "c": {"d": 4}}),
    ],
)
def test_remove_none_cases(get_topo_inputs_config, tmp_path, input_data, expected):
    """
    Test remove_none from dictionary
    """
    user_config = get_topo_inputs_config
    user_config["outputs"] = {"path": str(tmp_path)}
    workflows = Topo(user_config)
    assert workflows.remove_none(input_data) == expected


def test_load_config(get_topo_inputs_config, tmp_path):
    """
    Test load_config function
    """
    cfg = get_topo_inputs_config

    # Succeed
    with open(tmp_path / "temp_config.yaml", "w", encoding="utf-8") as f:
        yaml.dump(cfg, f, allow_unicode=True, default_flow_style=False)

    workflows = Topo(str(tmp_path / "temp_config.yaml"))
    assert workflows.load_config() == cfg

    # Fail
    with pytest.raises(FileNotFoundError, match=f"{tmp_path}/tempconfig.yaml does not exist"):
        _ = Topo(str(tmp_path / "tempconfig.yaml"))


def test_generate_graph(get_topo_inputs_config, tmp_path):
    """
    Test generate_plot function
    """
    dem = xdem.DEM(xdem.examples.get_path("longyearbyen_tba_dem"))
    title = "test_generate_graph"

    user_config = get_topo_inputs_config
    user_config["outputs"] = {"path": str(tmp_path)}
    workflows = Topo(user_config)

    workflows.generate_plot(dem, title)
    out = tmp_path / "plots" / f"{title}.png"
    assert out.exists()


@pytest.mark.parametrize(
    "inputs, expected",
    [
        pytest.param({"a": 1.2345, "b": 2.9876}, {"a": 1.23, "b": 2.99}, id="test_flat_dict"),
        pytest.param(
            {"a": {"b": 3.14159, "c": [1.6666, 2.5555]}}, {"a": {"b": 3.14, "c": [1.67, 2.56]}}, id="test_nested_dict"
        ),
        pytest.param({"a": (1.23456, 9.8765)}, {"a": (1.23, 9.88)}, id="test_tuple_values"),
        pytest.param(
            {"a": 5.678, "b": "string", "c": 42, "d": [3.3333, "text"]},
            {"a": 5.68, "b": "string", "c": 42, "d": [3.33, "text"]},
            id="test_mixed_types",
        ),
        pytest.param({"a": np.float64(2.71828)}, {"a": 2.72}, id="test_numpy_float"),
    ],
)
def test_floats_process(get_topo_inputs_config, tmp_path, inputs, expected):
    """
    Test floats_process function
    """
    user_config = get_topo_inputs_config
    user_config["outputs"] = {"path": str(tmp_path)}
    workflows = Topo(user_config)

    assert workflows.floats_process(inputs) == expected


def test_save_stat_as_csv(get_topo_inputs_config, tmp_path):
    """
    Test save_stat_as_csv function
    """
    user_config = get_topo_inputs_config
    user_config["outputs"] = {"path": str(tmp_path)}
    workflows = Topo(user_config)

    data = {"a": 1.2345, "b": 2.9876}
    title = "test_save_stat_as_csv"
    workflows.save_stat_as_csv(data, title)
    out = tmp_path / "tables" / f"{title}_stats.csv"
    assert out.exists()

    with open(out, newline="", encoding="utf-8") as f:
        reader = csv.DictReader(f)
        final_dict = list(reader)

    assert final_dict == [{"a": "1.2345", "b": "2.9876"}]


@pytest.mark.skip(reason="not implemented")
def test_load_dem():
    """
    Test generate_dem function
    """



# Copyright (c) 2025 Centre National d'Etudes Spatiales (CNES).
#
# This file is part of the xDEM project:
# https://github.com/glaciohack/xdem
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Topo class from workflows.
"""
import logging
import math
from pathlib import Path
from typing import Any, Dict

import matplotlib.pyplot as plt
import yaml  # type: ignore

import xdem
from xdem.workflows.schemas import TOPO_SCHEMA
from xdem.workflows.workflows import Workflows


class Topo(Workflows):
    """
    Topo class from workflows.
    """

    def __init__(self, config_dem: str | Dict[str, Any]):
        """
        Initialize Topo class
        :param config_dem: Path to a user configuration file
        """

        self.schema = TOPO_SCHEMA

        super().__init__(config_dem)

        self.dem, self.inlier_mask, path_to_mask = self.load_dem(self.config["inputs"]["reference_elev"])
        self.generate_plot(self.dem, "elev_map", cmap="terrain", cbar_title="Elevation (m)")

        if self.inlier_mask is not None:
            self.generate_plot(
                self.dem,
                "masked_elev_map",
                mask_path=path_to_mask,
                cmap="terrain",
                cbar_title="Elevation (m)",
            )

        self.config_attributes = self.config["terrain_attributes"]
        if isinstance(self.config_attributes, dict):
            self.list_attributes = list(self.config_attributes.keys())
        else:
            self.list_attributes = self.config_attributes

        yaml_str = yaml.dump(self.config, allow_unicode=True, Dumper=self.NoAliasDumper)
        Path(self.outputs_folder / "used_config.yaml").write_text(yaml_str, encoding="utf-8")

        self.config = self.remove_none(self.config)  # type: ignore

    def generate_terrain_attributes_tiff(self) -> None:
        """
        Generate terrain attributes tiff
        """

        attribute_extra = {}

        from_str_to_fun = {
            "slope": lambda: self.dem.slope(**attribute_extra),
            "aspect": lambda: self.dem.aspect(**attribute_extra),
            "hillshade": lambda: self.dem.hillshade(**attribute_extra),
            "curvature": lambda: self.dem.curvature(**attribute_extra),
            "planform_curvature": lambda: self.dem.planform_curvature(**attribute_extra),
            "profile_curvature": lambda: self.dem.profile_curvature(**attribute_extra),
            "maximum_curvature": lambda: self.dem.maximum_curvature(**attribute_extra),
            "topographic_position_index": lambda: self.dem.topographic_position_index(**attribute_extra),
            "terrain_ruggedness_index": lambda: self.dem.terrain_ruggedness_index(**attribute_extra),
            "roughness": lambda: self.dem.roughness(**attribute_extra),
            "rugosity": lambda: self.dem.rugosity(**attribute_extra),
            "fractal_roughness": lambda: self.dem.fractal_roughness(**attribute_extra),
        }
        for attr in self.list_attributes:
            if isinstance(self.config_attributes, dict):
                attribute_extra = self.config_attributes.get(attr).get("extra_information", {})  # type: ignore
            attribute = from_str_to_fun[attr]()
            logging.info(f"Saving {attr} as a raster file ({attr}.tif)")
            attribute.save(self.outputs_folder / "rasters" / f"{attr}.tif")

    def generate_terrain_attributes_png(self) -> None:
        """
        Generates an image png containing the plots of the terrain attributes requested by the user.
        :return: None
        """

        logging.info(f"Computing attributes : {self.list_attributes}")

        attributes = xdem.terrain.get_terrain_attribute(
            self.dem.data,
            resolution=self.dem.res,
            attribute=self.list_attributes,
        )

        n = len(attributes)

        ncols = 2
        nrows = math.ceil(n / ncols)

        plt.figure(figsize=(8, 6.5))

        plt_extent = [self.dem.bounds.left, self.dem.bounds.right, self.dem.bounds.bottom, self.dem.bounds.top]

        attribute_params = {
            "hillshade": {"label": "Hillshade", "cmap": "Greys_r", "vlim": (None, None)},
            "slope": {"label": "Slope (°)", "cmap": "Reds", "vlim": (None, None)},
            "aspect": {"label": "Aspect (°)", "cmap": "twilight", "vlim": (None, None)},
            "curvature": {"label": "Curvature (100 / m)", "cmap": "RdGy_r", "vlim": (-2, 2)},
            "planform_curvature": {"label": "Planform curvature (100 / m)", "cmap": "RdGy_r", "vlim": (-2, 2)},
            "profile_curvature": {"label": "Profile curvature (100 / m)", "cmap": "RdGy_r", "vlim": (-2, 2)},
            "terrain_ruggedness_index": {"label": "Terrain Ruggedness Index", "cmap": "Purples", "vlim": (None, None)},
            "rugosity": {"label": "Rugosity", "cmap": "YlOrRd", "vlim": (None, None)},
            "topographic_position_index": {
                "label": "Topographic position index (m)",
                "cmap": "Spectral",
                "vlim": (None, None),
            },
            "roughness": {"label": "Roughness (m)", "cmap": "Oranges", "vlim": (None, None)},
            "fractal_dimension": {"label": "Fractal roughness (dimensions)", "cmap": "Reds", "vlim": (None, None)},
        }

        for i, attr in enumerate(self.list_attributes):
            plt.subplot(nrows, ncols, i + 1)

            params = attribute_params.get(attr, {})
            cmap = params.get("cmap", "viridis")
            label = params.get("label", f"Attribute {i + 1}")
            vmin, vmax = params.get("vlim", (None, None))

            plt.imshow(attributes[i].squeeze(), cmap=cmap, extent=plt_extent, vmin=vmin, vmax=vmax)
            cbar = plt.colorbar()
            cbar.set_label(label)
            plt.xticks([])
            plt.yticks([])

        plt.tight_layout()
        plt.savefig(self.outputs_folder / "plots" / "terrain_attributes_map.png")
        plt.close()

    def run(self) -> None:
        """
        Run function for the coregistration workflow
        :return: None
        """

        # Global information
        dem_informations = {
            "Driver": self.dem.driver,
            "Filename": self.dem.filename,
            "Grid size": self.dem.vcrs_grid,
            "Number of band": self.dem.bands,
            "Data types": self.dem.dtype,
            "Nodata Value": self.dem.nodata,
            "Pixel interpretation": self.dem.area_or_point,
            "Pixel size": self.dem.res,
            "Width": self.dem.width,
            "Height": self.dem.height,
            "Transform": self.dem.transform,
            "Bounds": self.dem.bounds,
        }

        # Statistics
        list_metrics = self.config["statistics"]
        if list_metrics is not None:
            stats_dem = self.dem.get_stats(list_metrics)
            self.save_stat_as_csv(stats_dem, "stats_elev")
            stats_dem_mask = self.dem.get_stats(list_metrics, inlier_mask=self.inlier_mask)
            if self.inlier_mask is not None:
                self.save_stat_as_csv(stats_dem_mask, "stats_elev_mask")
            logging.info(f"Computing metrics on reference elevation: {list_metrics}")

        # Terrain attributes
        if self.list_attributes is not None:
            self.generate_terrain_attributes_png()
            if self.level > 1:
                self.generate_terrain_attributes_tiff()
        else:
            logging.info("Computing terrain attributes: None")

        # Generate HTML
        self.dico_to_show.append(("Elevation information", dem_informations))
        if list_metrics is not None:
            self.dico_to_show.append(("Global statistics", self.floats_process(stats_dem)))
        if self.inlier_mask is not None:
            self.dico_to_show.append(("Mask statistics", self.floats_process(stats_dem_mask)))

        self.create_html(self.dico_to_show)

        # Remove empty folder
        for folder in self.outputs_folder.rglob("*"):
            if folder.is_dir():
                try:
                    folder.rmdir()
                except OSError:
                    pass

    def create_html(self, list_dict: list[tuple[str, dict[str, Any]]]) -> None:
        """
        Create HTML page from png files and table
        :param list_dict: list containing tuples of title and various dictionaries
        :return: None
        """

        html = "<html>\n<head><meta charset='UTF-8'><title>Topographic summary results</title></head>\n<body>\n"

        html += "<h2>Elevation Model</h2>\n"
        html += "<img src='plots/elev_map.png' alt='Image PNG' style='max-width: 100%; height: auto;'>\n"

        if self.inlier_mask is not None:
            html += "<h2>Masked elevation Model</h2>\n"
            html += "<img src='plots/masked_elevation.png' alt='Image PNG' style='max-width: 100%; height: auto;'>\n"

        for title, dictionary in list_dict:
            html += "<div style='clear: both; margin-bottom: 30px;'>\n"  # type: ignore
            html += f"<h2>{title}</h2>\n"
            html += "<table border='1' cellspacing='0' cellpadding='5'>\n"
            html += "<tr><th>Information</th><th>Value</th></tr>\n"
            for key, value in dictionary.items():
                html += f"<tr><td>{key}</td><td>{value}</td></tr>\n"
            html += "</table>\n"
            html += "</div>\n"

        html += "<h2>Terrain attributes</h2>\n"
        html += "<img src='plots/terrain_attributes_map.png' alt='Image PNG' style='max-width: 100%; height: auto;'>\n"

        html += "</body>\n</html>"

        with open(self.outputs_folder / "report.html", "w", encoding="utf-8") as f:
            f.write(html)



# Copyright (c) 2025 Centre National d'Etudes Spatiales (CNES).
#
# This file is part of the xDEM project:
# https://github.com/glaciohack/xdem
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Accuracy class from workflow
"""
import logging
from pathlib import Path
from typing import Any, Dict

import matplotlib.pyplot as plt
import numpy as np
import yaml  # type: ignore
from geoutils.raster import RasterType
from numpy import floating

import xdem
from xdem.workflows.schemas import ACCURACY_SCHEMA
from xdem.workflows.workflows import Workflows


class Accuracy(Workflows):
    """
    Accuracy class from workflow
    """

    def __init__(self, config_dem: str | Dict[str, Any]) -> None:
        """
        Initialize the Accuracy class
        :param config_dem: Path to a user configuration file
        """

        self.schema = ACCURACY_SCHEMA

        super().__init__(config_dem)

        self.to_be_aligned_elev, tba_mask, tba_path_mask = self.load_dem(self.config["inputs"]["to_be_aligned_elev"])
        self.reference_elev, ref_mask, ref_mask_path = self.load_dem(self.config["inputs"]["reference_elev"])
        if self.reference_elev is None:
            self.reference_elev = self._get_reference_elevation()
        self.generate_plot(self.reference_elev, "reference_elev_map")
        self.generate_plot(self.to_be_aligned_elev, "to_be_aligned_elev_map")

        self.inlier_mask = None
        if ref_mask is not None and tba_mask is not None:
            self.inlier_mask = tba_mask
            path_mask = tba_path_mask
        else:
            self.inlier_mask = ref_mask or tba_mask
            path_mask = ref_mask_path or tba_path_mask

        if self.inlier_mask is not None:
            self.generate_plot(
                self.to_be_aligned_elev,
                "masked_elev_map",
                mask_path=path_mask,
                cmap="terrain",
                cbar_title="Elevation (m)",
            )

        self.compute_coreg = self.config["coregistration"]["process"]

        if not self.compute_coreg:
            del self.config["coregistration"]["sampling_grid"]
            del self.config["coregistration"]["step_one"]

        yaml_str = yaml.dump(self.config, allow_unicode=True, Dumper=self.NoAliasDumper)
        Path(self.outputs_folder / "used_config.yaml").write_text(yaml_str, encoding="utf-8")

        self.config = self.remove_none(self.config)  # type: ignore

    def _get_reference_elevation(self) -> float:
        """
        Get reference elevation
        """

        raise NotImplementedError("For now it doesn't working, please add a reference DEM")

    def _compute_coregistration(self) -> RasterType:
        """
        Wrapper for coregistration
        """

        coreg_extra = {}

        # Coregister
        from_str_to_fun = {
            "NuthKaab": lambda: xdem.coreg.NuthKaab(**coreg_extra),
            "DhMinimize": lambda: xdem.coreg.DhMinimize(**coreg_extra),
            "VerticalShift": lambda: xdem.coreg.VerticalShift(**coreg_extra),
            "DirectionalBias": lambda: xdem.coreg.DirectionalBias(**coreg_extra),
            "TerrainBias": lambda: xdem.coreg.TerrainBias(**coreg_extra),
            "LZD": lambda: xdem.coreg.LZD(**coreg_extra),
        }

        coreg_steps = ["step_one", "step_two", "step_three"]
        coreg_functions = []

        for step in coreg_steps:
            config_coreg = self.config["coregistration"].get(step)
            if config_coreg:
                method_name = config_coreg.get("method")
                coreg_extra = config_coreg.get("extra_information", {})
                coreg_fun = from_str_to_fun[method_name]()
                coreg_functions.append(coreg_fun)

        my_coreg = sum(coreg_functions[1:], coreg_functions[0]) if len(coreg_functions) > 1 else coreg_functions[0]

        # Coregister
        aligned_elev = self.to_be_aligned_elev.coregister_3d(self.reference_elev, my_coreg, self.inlier_mask)
        aligned_elev.save(self.outputs_folder / "rasters" / "aligned_elev.tif")

        self.dico_to_show.append(("Coregistration user configuration", self.config["coregistration"]))

        for idx, step in enumerate(coreg_steps):
            config_coreg = self.config["coregistration"].get(step)
            if config_coreg:
                method_name = config_coreg.get("method")
                self.dico_to_show.append(
                    (f"{method_name} inputs", self.floats_process(coreg_functions[idx].meta["inputs"]))
                )
                self.dico_to_show.append(
                    (f"{method_name} outputs", self.floats_process(coreg_functions[idx].meta["outputs"]))
                )

        return aligned_elev

    def _compute_reproj(self, test_dem: str) -> None:
        """
        Compute reprojection
        :param test_dem: str value for testing the target dem
        """
        # Reproject data

        sampling = self.config["coregistration"]["sampling_grid"]
        if sampling == test_dem:
            return  # No reprojection needed

        logging.info(f"Computing reprojection on {test_dem}")

        if sampling == "reference_elev":
            src, target = self.to_be_aligned_elev, self.reference_elev
            name = "to_be_aligned_elev"
            reprojected = src.reproject(target, silent=True)
            self.to_be_aligned_elev = reprojected
        elif sampling == "to_be_aligned_elev":
            src, target = self.reference_elev, self.to_be_aligned_elev
            name = "reference_elevation"
            reprojected = src.reproject(target, silent=True)
            self.reference_elev = reprojected

        if self.level > 1:
            output_path = self.outputs_folder / "rasters" / f"{name}_reprojected.tif"
            reprojected.save(output_path)

    def _get_stats(self, dem: RasterType, name_of_data: str = "") -> floating[Any] | dict[str, floating[Any]]:
        """
        Return a list of computed statistics chose by user or the default one.
        :param dem: DEM to process
        :name_of_data: string for logging
        """
        # Compute user statistics
        dict_stats_aliased = {}
        list_to_compute = self.config["statistics"]
        if list_to_compute is not None:
            logging.info(f"Computing statistics on {name_of_data}: {list_to_compute}")
            dict_stats = dem.get_stats(list_to_compute)

            # Aliases for nicer CSV headers
            aliases = {
                "mean": "Mean",
                "median": "Median",
                "max": "Maximum",
                "min": "Minimum",
                "sum": "Sum",
                "sumofsquares": "Sum of squares",
                "90thpercentile": "90th percentile",
                "le90": "LE90",
                "nmad": "NMAD",
                "rmse": "RMSE",
                "std": "STD",
                "standarddeviation": "Standard deviation",
                "validcount": "Valid count",
                "totalcount": "Total count",
                "percentagevalidpoints": "Percentage valid points",
            }

            dict_stats_aliased = {aliases.get(k, k): v for k, v in dict_stats.items()}

        return dict_stats_aliased

    def _compute_histogram(self) -> None:
        """
        Compute altitudes difference histogram.
        """
        logging.info("Computing histogram on altitude difference")
        plt.figure(figsize=(7, 8))
        bins = np.linspace(self.stats_before["min"], self.stats_before["max"], 300)
        plt.hist(self.diff_before.data.flatten(), bins=bins, color="g", alpha=0.6, label="Before_coregistration")
        plt.hist(self.diff_after.data.flatten().data, bins=bins, color="b", alpha=0.6, label="After_coregistration")
        plt.text(
            -50,
            50000,
            f'Before\nmed: {self.stats_before["median"]:.2f}\nnmad: {self.stats_before["nmad"]:.2f}',
            color="green",
            fontsize=12,
        )
        plt.text(
            30,
            50000,
            f'After\nmed: {self.stats_after["median"]:.2f}\nnmad: {self.stats_after["nmad"]:.2f}',
            color="blue",
            fontsize=12,
        )
        plt.title("Differences histograms", fontsize=14)
        plt.xlabel("Elev. Diff. (m)")
        plt.ylabel("Count (px)")
        plt.legend()
        plt.grid(False)
        plt.tight_layout()
        plt.savefig(self.outputs_folder / "plots" / "elev_diff_histo.png")
        plt.close()

    def run(self) -> None:
        """
        Run function for the coregistration workflow
        :return: None
        """

        if self.compute_coreg:
            # Reprojection step
            self._compute_reproj("reference_elev")
            self._compute_reproj("to_be_aligned_elev")

            # Coregistration step
            aligned_elev = self._compute_coregistration()
        else:
            logging.info("Coregistration not executed, returned to_be_aligned_elev")
            aligned_elev = self.to_be_aligned_elev

        output_grid = self.config["outputs"]["output_grid"]
        ref_elev = self.reference_elev if output_grid == "reference_elev" else self.to_be_aligned_elev

        vmin = vmax = None

        if self.compute_coreg:
            diff_pairs = [("before", self.to_be_aligned_elev), ("after", aligned_elev.reproject(ref_elev))]
        else:
            diff_pairs = [("", self.to_be_aligned_elev)]

        for label, dem in diff_pairs:
            diff = dem - ref_elev
            stats_keys = ["min", "max", "nmad", "median"]
            stats = diff.get_stats(stats_keys)

            if label == "before":
                self.diff_before, self.stats_before = diff, stats
                vmin, vmax = stats["min"], stats["max"]
            elif label == "after":
                self.diff_after, self.stats_after = diff, stats
            else:
                self.diff = diff
                vmin, vmax = stats["min"], stats["max"]

            suffix = f"_elev_{label}_coreg" if label else "_elev"
            self.generate_plot(diff, f"diff{suffix}", vmin=vmin, vmax=vmax, cmap="RdBu")

        if self.compute_coreg:
            stat_items = [
                (self.reference_elev, "reference_elev", "Statistics on reference elevation", 2),
                (self.to_be_aligned_elev, "to_be_aligned_elev", "Statistics on to be aligned elevation", 2),
                (aligned_elev, "aligned_elev", "Statistics on aligned elevation", 1),
                (
                    self.diff_before,
                    "diff_elev_before_coreg",
                    "Statistics on altitude difference before coregistration",
                    1,
                ),
                (self.diff_after, "diff_elev_after_coreg", "Statistics on altitude difference after coregistration", 1),
            ]
        else:
            stat_items = [
                (self.reference_elev, "reference_elev", "Statistics on reference elevation", 2),
                (self.to_be_aligned_elev, "to_be_aligned_elev", "Statistics on to be aligned elevation", 2),
                (self.diff, "diff_elev", "Statistics on altitude difference", 2),
            ]

        for data, fname, title, level in stat_items:
            stats = self._get_stats(data, fname)
            if (level <= self.level or not self.compute_coreg) and self.config["statistics"] is not None:
                self.save_stat_as_csv(stats, fname)
            self.dico_to_show.append((title, self.floats_process(stats)))

        if self.compute_coreg:
            self._compute_histogram()
            if self.level > 1:
                self.diff_before.save(self.outputs_folder / "rasters" / "diff_elev_before_coreg.tif")
                self.diff_after.save(self.outputs_folder / "rasters" / "diff_elev_after_coreg.tif")

        self.create_html(self.dico_to_show)

        # Remove empty folder
        for folder in self.outputs_folder.rglob("*"):
            if folder.is_dir():
                try:
                    folder.rmdir()
                except OSError:
                    pass

    def create_html(self, list_dict: list[tuple[str, dict[str, Any]]]) -> None:
        """
        Create HTML page from png files and table
        :param list_dict: list containing tuples of title and various dictionaries
        :return: None
        """
        html = "<html>\n<head><meta charset='UTF-8'><title>Qualify elevation results</title></head>\n<body>\n"

        html += "<h2>Digital Elevation Model</h2>\n"
        html += "<div style='display: flex; gap: 10px;'>\n"
        html += (
            "  <img src='plots/reference_elev_map.png' alt='Image PNG' "
            "style='max-width: 100%; height: auto; width: 40%;'>\n"
        )
        html += (
            "  <img src='plots/to_be_aligned_elev_map.png' alt='Image PNG' style='max-width: "
            "100%; height: auto; width: 40%;'>\n"
        )
        html += "</div>\n"

        for title, dictionary in list_dict:  # type: ignore
            html += "<div style='clear: both; margin-bottom: 30px;'>\n"
            html += f"<h2>{title}</h2>\n"
            html += "<table border='1' cellspacing='0' cellpadding='5'>\n"
            html += "<tr><th>Information</th><th>Value</th></tr>\n"
            for key, value in dictionary.items():
                html += f"<tr><td>{key}</td><td>{value}</td></tr>\n"
            html += "</table>\n"
            html += "</div>\n"

        if self.compute_coreg:
            html += "<h2>Altitude differences</h2>\n"
            html += "<div style='display: flex; gap: 10px;'>\n"
            html += (
                "  <img src='plots/diff_elev_before_coreg.png' alt='Image PNG' style='max-width: "
                "40%; height: auto; width: 50%;'>\n"
            )
            html += (
                "  <img src='plots/diff_elev_after_coreg.png' alt='Image PNG' style='max-width: "
                "40%; height: auto; width: 50%;'>\n"
            )
            html += "</div>\n"

            html += "<h2>Differences histogram</h2>\n"
            html += "<img src='plots/elev_diff_histo.png' alt='Image PNG' style='max-width: 40%; height: auto;'>\n"

        else:
            html += "<h2>Altitude differences</h2>\n"
            html += "<div style='display: flex; gap: 10px;'>\n"
            html += (
                "  <img src='plots/diff_elev.png' alt='Image PNG' style='max-width: "
                "40%; height: auto; width: 50%;'>\n"
            )
            html += "</div>\n"

        html += """
             </div>
         </body>
         </html>
         """

        with open(self.outputs_folder / "report.html", "w", encoding="utf-8") as f:
            f.write(html)



# Copyright (c) 2025 Centre National d'Etudes Spatiales (CNES).
#
# This file is part of the xDEM project:
# https://github.com/glaciohack/xdem
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Fixtures for test_workflows
"""
# mypy: disable-error-code=no-untyped-def

import pytest

import xdem
from xdem.workflows import Accuracy
from xdem.workflows.schemas import TERRAIN_ATTRIBUTES_DEFAULT


@pytest.fixture()
def get_topo_inputs_config():
    """
    Return minimal configuration for inputs in topo
    """
    return {
        "inputs": {
            "reference_elev": {
                "path_to_elev": xdem.examples.get_path("longyearbyen_tba_dem"),
                "path_to_mask": xdem.examples.get_path("longyearbyen_glacier_outlines"),
            }
        },
    }


@pytest.fixture()
def get_accuracy_inputs_config():
    """
    Return minimal configuration for inputs in accuracy
    """
    return {
        "inputs": {
            "reference_elev": {
                "path_to_elev": xdem.examples.get_path("longyearbyen_ref_dem"),
                "path_to_mask": xdem.examples.get_path("longyearbyen_glacier_outlines"),
            },
            "to_be_aligned_elev": {
                "path_to_elev": xdem.examples.get_path("longyearbyen_tba_dem"),
                "path_to_mask": xdem.examples.get_path("longyearbyen_glacier_outlines"),
            },
        },
    }


@pytest.fixture()
def get_accuracy_object_with_run(tmp_path):
    """
    Generate classical accuracy object
    """
    user_config = {
        "inputs": {
            "reference_elev": {
                "path_to_elev": xdem.examples.get_path("longyearbyen_ref_dem"),
                "path_to_mask": xdem.examples.get_path("longyearbyen_glacier_outlines"),
            },
            "to_be_aligned_elev": {
                "path_to_elev": xdem.examples.get_path("longyearbyen_tba_dem"),
                "path_to_mask": xdem.examples.get_path("longyearbyen_glacier_outlines"),
            },
        },
        "outputs": {"path": str(tmp_path)},
    }
    workflows = Accuracy(user_config)
    workflows.run()

    return workflows


@pytest.fixture()
def pipeline_topo():
    """
    Return default configuration for pipeline topo_summary
    """
    return {
        "inputs": {
            "reference_elev": {
                "path_to_elev": xdem.examples.get_path("longyearbyen_tba_dem"),
                "path_to_mask": xdem.examples.get_path("longyearbyen_glacier_outlines"),
            }
        },
        "statistics": [
            "mean",
            "median",
            "max",
            "min",
            "sum",
            "sumofsquares",
            "90thpercentile",
            "le90",
            "nmad",
            "rmse",
            "std",
            "standarddeviation",
            "validcount",
            "totalcount",
            "percentagevalidpoints",
        ],
        "terrain_attributes": TERRAIN_ATTRIBUTES_DEFAULT,
        "outputs": {"path": "outputs", "level": 1},
    }


@pytest.fixture()
def list_default_terrain_attributes():
    """
    Return default list of terrain attributes
    """
    return TERRAIN_ATTRIBUTES_DEFAULT



"""Functions to test the example data."""

from __future__ import annotations

import geoutils as gu
import numpy as np
import pytest
from geoutils import Raster, Vector

from xdem import examples
from xdem._typing import NDArrayf


def load_examples() -> tuple[Raster, Raster, Vector, Raster]:
    """Load example files to try coregistration methods with."""

    ref_dem = Raster(examples.get_path("longyearbyen_ref_dem"))
    tba_dem = Raster(examples.get_path("longyearbyen_tba_dem"))
    glacier_mask = Vector(examples.get_path("longyearbyen_glacier_outlines"))
    ddem = Raster(examples.get_path("longyearbyen_ddem"))

    return ref_dem, tba_dem, glacier_mask, ddem


class TestExamples:

    ref_dem, tba_dem, glacier_mask, ddem = load_examples()

    @pytest.mark.parametrize(
        "rst_truevals_abs",
        [
            (ref_dem, np.array([465.11816, 207.3236, 208.30563, 748.7337, 797.28644], dtype=np.float32), None),
            (tba_dem, np.array([464.6715, 213.7554, 207.8788, 760.8192, 797.3268], dtype=np.float32), None),
            (ddem, np.array([1.37, -1.67, 0.13, -10.10, 2.49], dtype=np.float32), 10e-3),
        ],
    )  # type: ignore
    def test_array_content(self, rst_truevals_abs: tuple[Raster, NDArrayf, float]) -> None:
        """Let's ensure the data arrays in the examples are always the same by checking randomly some values"""

        rst = rst_truevals_abs[0]
        truevals = rst_truevals_abs[1]
        abs = rst_truevals_abs[2]

        rng = np.random.default_rng(42)
        values = rng.choice(rst.data.data.flatten(), size=5, replace=False)

        assert values == pytest.approx(truevals, abs=abs)

    # Note: Following PR #329, no gaps on DEM edges after coregistration
    @pytest.mark.parametrize("rst_and_truenodata", [(ref_dem, 0), (tba_dem, 0), (ddem, 0)])  # type: ignore
    def test_array_nodata(self, rst_and_truenodata: tuple[Raster, int]) -> None:
        """Let's also check that the data arrays have always the same number of not finite values"""

        rst = rst_and_truenodata[0]
        truenodata = rst_and_truenodata[1]
        mask = gu.raster.get_array_and_mask(rst)[1]

        assert np.sum(mask) == truenodata



# Copyright (c) 2025 Centre National d'Etudes Spatiales (CNES).
#
# This file is part of the xDEM project:
# https://github.com/glaciohack/xdem
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Workflow class
"""

import csv
import logging
import os
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Dict, List, Union

import geoutils as gu
import matplotlib.pyplot as plt
import numpy as np
import yaml  # type: ignore
from geoutils import Raster
from geoutils.raster import RasterType
from yaml.dumper import SafeDumper  # type: ignore

import xdem
from xdem import DEM
from xdem.coreg.base import InputCoregDict, OutputCoregDict
from xdem.workflows.schemas import validate_configuration


class Workflows(ABC):
    """
    Abstract Class for workflows
    """

    def __init__(self, user_config: str | Dict[str, Any]) -> None:
        """
        Initialize the workflows class
        :param user_config: str path to a config file or dict as config
        :return: None
        """

        # Load configuration
        if isinstance(user_config, str):
            if not os.path.isfile(user_config):
                raise FileNotFoundError(f"{user_config} does not exist")
            self.config_path = user_config
            config_not_verify = self.load_config()
        elif isinstance(user_config, dict):
            config_not_verify = user_config
        else:
            raise ValueError(
                "The configuration should be provided either as a path to the configuration file"
                " or as a dictionary containing the configuration details."
            )

        self.config = validate_configuration(config_not_verify, self.schema)
        self.level = self.config["outputs"]["level"]

        self.outputs_folder = Path(self.config["outputs"]["path"])
        self.outputs_folder.mkdir(parents=True, exist_ok=True)

        for folder in ["plots", "rasters", "tables"]:
            Path(self.outputs_folder / folder).mkdir(parents=True, exist_ok=True)

        self.dico_to_show = [
            ("Information about inputs", self.config["inputs"]),
        ]

    class NoAliasDumper(SafeDumper):  # type: ignore
        """
        NoAliasDumper to avoid id in YAML file
        """

        def ignore_aliases(self, data: Any) -> bool:
            """
            avoid id in YAML file
            """
            return True

    def load_config(self) -> Dict[str, Any]:
        """
        Load a configuration file
        :return: Configuration dictionary
        """
        if not os.path.exists(self.config_path):
            raise FileNotFoundError(f"File not found : {self.config_path}")
        with open(self.config_path) as f:
            return yaml.safe_load(f)

    def generate_plot(self, dem: RasterType, title: str, mask_path: str = None, **kwargs: Any) -> None:
        """
        Generate plot from a DEM
        :param dem: Digital Elevation model
        :param title: title of graph
        :param mask_path: Path to mask
        :return: None
        """

        plot_title = title.replace("_", " ")

        if mask_path is None:
            dem.plot(title=plot_title, **kwargs)
            plt.savefig(self.outputs_folder / "plots" / f"{title}.png")
            plt.close()
        else:
            mask = gu.Vector(mask_path)
            mask = mask.crop(dem)
            dem.plot(title=plot_title, **kwargs)
            mask.plot(dem, ec="k", fc="none")
            plt.savefig(self.outputs_folder / "plots" / f"{title}.png")
            plt.close()

    def floats_process(
        self, dict_with_floats: Dict[str, Any] | InputCoregDict | OutputCoregDict | Any
    ) -> Dict[str, Any]:  # type: ignore
        """
        Allows rounding all floats present in a dictionary to two decimal places
        :param dict_with_floats: Dictionary with float
        :return: Dictionary with floats
        """
        if isinstance(dict_with_floats, dict):
            return {k: self.floats_process(v) for k, v in dict_with_floats.items()}
        elif isinstance(dict_with_floats, list):
            return [self.floats_process(elem) for elem in dict_with_floats]  # type: ignore
        elif isinstance(dict_with_floats, tuple):
            return tuple(self.floats_process(elem) for elem in dict_with_floats)  # type: ignore
        elif isinstance(dict_with_floats, (float, np.floating)):
            return round(float(dict_with_floats), 2)  # type: ignore
        else:
            return dict_with_floats

    @staticmethod
    def load_dem(config_dem: Dict[str, Any] | None) -> tuple[DEM, Raster, str | None]:
        """
        Generate DEM from user configuration dictionary
        :param config_dem: Configuration dictionary
        :return: DEM
        """
        mask_path = None
        if config_dem is not None:
            dem = xdem.DEM(config_dem["path_to_elev"])
            inlier_mask = None
            from_vcrs = config_dem["from_vcrs"]
            to_vcrs = config_dem["to_vcrs"]
            if from_vcrs:
                dem.set_vcrs(from_vcrs)
            if to_vcrs:
                if from_vcrs != to_vcrs:
                    dem.to_vcrs(to_vcrs)
            if config_dem.get("force_source_nodata") is not None:
                dem.set_nodata(config_dem["force_source_nodata"])
            if config_dem.get("path_to_mask") is not None:
                mask_path = config_dem["path_to_mask"]
                mask = gu.Vector(mask_path)
                inlier_mask = ~mask.create_mask(dem)

            return dem, inlier_mask, mask_path
        else:
            logging.warning("No DEM provided")
            return None, None, None  # type: ignore

    def remove_none(self, dico: Union[Dict[str, Any], List[Any]]) -> Union[Dict[str, Any], List[Any]]:
        """
        Recursively remove all keys whose values are None from a dictionary.
        :param dico: dictionary to clean
        :return: cleaned dictionary
        """
        if isinstance(dico, dict):
            return {
                k: self.remove_none(v) for k, v in dico.items() if v is not None and self.remove_none(v) is not None
            }
        elif isinstance(dico, list):
            cleaned_list = [self.remove_none(v) for v in dico if v is not None]
            return [v for v in cleaned_list if v is not None]
        else:
            return dico

    @abstractmethod
    def create_html(self, list_dict: list[tuple[str, dict[str, Any]]]) -> None:
        """
        Create HTML page from png files and table
        :param list_dict: list containing tuples of title and various dictionaries
        :return: None
        """

    def save_stat_as_csv(self, data: dict[str, float], file_name: str) -> None:
        """
        Save the statistics into a CSV file
        :param data: Statistics dictionary
        :param file_name: Name of csv file
        """
        cleaned_data = {k: float(v) if isinstance(v, (np.float32, np.float64)) else v for k, v in data.items()}

        fieldnames = list(cleaned_data.keys())
        filename = self.outputs_folder / "tables" / f"{file_name}_stats.csv"

        with filename.open(mode="w", newline="", encoding="utf-8") as csvfile:
            writer = csv.DictWriter(csvfile, fieldnames=fieldnames)
            writer.writeheader()
            writer.writerow(cleaned_data)



# Copyright (c) 2025 Centre National d'Etudes Spatiales (CNES).
#
# This file is part of the xDEM project:
# https://github.com/glaciohack/xdem
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
test for schema files
"""
import pyproj

# mypy: disable-error-code=no-untyped-def
import pytest

from xdem.workflows import schemas


def test_validate_base_configuration(get_topo_inputs_config, get_accuracy_inputs_config):
    """
    Test validate_base_configuration function
    """
    schemas.validate_configuration(get_topo_inputs_config, schemas.TOPO_SCHEMA)
    schemas.validate_configuration(get_accuracy_inputs_config, schemas.ACCURACY_SCHEMA)


def test_wrong_path(get_topo_inputs_config):
    """
    Test wrong_path function
    """
    info_conf = get_topo_inputs_config
    info_conf["inputs"]["reference_elev"]["path_to_elev"] = "doesn_t_exist.tif"

    with pytest.raises(ValueError, match="Path does not exist: doesn_t_exist.tif"):
        _ = schemas.validate_configuration(info_conf, schemas.TOPO_SCHEMA)


@pytest.mark.parametrize(
    "new_param_config, expected",
    [
        pytest.param({"outputs": {"level": 3}}, "unallowed value 3", id="outputs_level"),
        pytest.param({"outputs": {"path": 3}}, "must be of string type", id="outputs_path_int"),
        pytest.param(
            {"outputs": {"level": "level_3"}},
            "must be of integer type",
            id="outputs_level_str",
        ),
        pytest.param({"statistics": {"mean": 0.5}}, "must be of list type", id="statistics_dict_in"),
        pytest.param(
            {"statistics": ["wrong_metrics"]},
            "unallowed values",
            id="statistics_wrong_metrics",
        ),
        pytest.param(
            {"terrain_attributes": ["wrong_attr"]},
            "no definitions validate",
            id="terrain_attributes_wrong_attr",
        ),
        pytest.param(
            {"terrain_attributes": {"slope": {"extra_information": 2}}},
            "must be of dict type",
            id="terrain_attributes_dict_attr",
        ),
    ],
)
def test_validate_info_configuration_with_errors(get_topo_inputs_config, new_param_config, expected):
    """
    Test validation of configuration with errors
    """
    info_conf = get_topo_inputs_config
    info_conf.update(new_param_config)

    with pytest.raises(ValueError, match=expected):
        _ = schemas.validate_configuration(info_conf, schemas.TOPO_SCHEMA)


@pytest.mark.parametrize(
    "new_param_config, expected",
    [
        pytest.param(
            {"coregistration": {"sampling_grid": 3}},
            "must be of string type",
            id="sampling_grid",
        ),
        pytest.param(
            {"coregistration": {"sampling_grid": "not_a_dem"}},
            "unallowed value not_a_dem",
            id="sampling_grid",
        ),
    ],
)
def test_validate_info_coreg_configuration_with_errors(get_accuracy_inputs_config, new_param_config, expected):
    """
    Test validation of coregistration configuration with errors
    """
    info_conf = get_accuracy_inputs_config
    info_conf.update(new_param_config)

    with pytest.raises(ValueError, match=expected):
        _ = schemas.validate_configuration(info_conf, schemas.ACCURACY_SCHEMA)


@pytest.mark.parametrize(
    "required,expected_required",
    [
        (True, True),
        (False, False),
    ],
)
def test_required_flag(required, expected_required):
    """
    Test required_flag in coregistration
    """
    schema = schemas.make_coreg_step(required=required)
    assert schema["required"] == expected_required
    assert schema["schema"]["method"]["required"] == expected_required


@pytest.mark.parametrize(
    "default_method,expected_present",
    [
        ("method_a", True),
        (None, False),
    ],
)
def test_default_method_handling(default_method, expected_present):
    """
    Test default method handling for coregistration
    """
    schema = schemas.make_coreg_step(default_method=default_method)
    assert ("default" in schema) == expected_present
    if expected_present:
        assert schema["default"]["method"] == default_method


def test_allowed_methods():
    """
    Test allowed_methods in coregistration
    """
    schema = schemas.make_coreg_step()
    assert schema["schema"]["method"]["allowed"] == schemas.COREG_METHODS


def test_extra_information_is_optional():
    """
    Test extra_information_is_optional in coregistration
    """
    schema = schemas.make_coreg_step()
    assert "extra_information" in schema["schema"]
    assert not schema["schema"]["extra_information"]["required"]


@pytest.mark.parametrize(
    "prefix, vcrs",
    [
        ("from_vcrs", "EGM96"),
        ("from_vcrs", "EGM08"),
        ("from_vcrs", "Ellipsoid"),
        # ("from_vcrs", "no_kv_arcgp-2006-sk.tif"),
        ("from_vcrs", 4326),
    ],
)
def test_valid_from_vcrs(get_topo_inputs_config, pipeline_topo, prefix, vcrs):
    """
    Test valid VCRS function for 'from' and 'to'
    """
    info_conf = get_topo_inputs_config
    info_conf["inputs"]["reference_elev"].update({prefix: vcrs})

    pipeline_test = schemas.validate_configuration(info_conf, schemas.TOPO_SCHEMA)
    pipeline_topo["inputs"]["reference_elev"].update({prefix: vcrs})
    pipeline_topo["inputs"]["reference_elev"]["to_vcrs"] = None
    assert pipeline_topo == pipeline_test


@pytest.mark.parametrize(
    "prefix, vcrs",
    [
        ("to_vcrs", "EGM96"),
        ("to_vcrs", "EGM08"),
        ("to_vcrs", "Ellipsoid"),
        # ("to_vcrs", "no_kv_arcgp-2006-sk.tif"),
        ("to_vcrs", 4326),
    ],
)
def test_valid_to_vcrs(get_topo_inputs_config, pipeline_topo, prefix, vcrs):
    """
    Test valid VCRS function for 'from' and 'to'
    """
    info_conf = get_topo_inputs_config
    info_conf["inputs"]["reference_elev"].update({prefix: vcrs})

    pipeline_test = schemas.validate_configuration(info_conf, schemas.TOPO_SCHEMA)
    pipeline_topo["inputs"]["reference_elev"].update({prefix: vcrs})
    pipeline_topo["inputs"]["reference_elev"]["from_vcrs"] = None
    assert pipeline_topo == pipeline_test


@pytest.mark.parametrize(
    "wrong_vcrs, error",
    [
        pytest.param(
            "wrong",
            UserWarning,
            id="wrong_common",
        ),
        pytest.param(
            "wrong.txt",
            UserWarning,
            id="wrong_proj_grid",
        ),
        pytest.param(
            0000,
            pyproj.exceptions.CRSError,
            id="wrong_epsg_code",
        ),
    ],
)
def test_invalid_vcrs(get_topo_inputs_config, pipeline_topo, wrong_vcrs, error):
    """
    Test invalid crs
    """
    info_conf = get_topo_inputs_config
    info_conf["inputs"]["reference_elev"].update({"from_vcrs": wrong_vcrs})

    with pytest.raises(error):
        _ = schemas.validate_configuration(info_conf, schemas.TOPO_SCHEMA)


def test_topo_without_terrain_attributes_in_config(get_topo_inputs_config):
    """
    Test different value for terrain attributes in config
    """
    info_conf = get_topo_inputs_config
    doc = schemas.validate_configuration(info_conf, schemas.TOPO_SCHEMA)
    assert doc["terrain_attributes"] == schemas.TERRAIN_ATTRIBUTES_DEFAULT

    info_conf = get_topo_inputs_config
    info_conf["terrain_attributes"] = []
    doc = schemas.validate_configuration(info_conf, schemas.TOPO_SCHEMA)
    assert doc["terrain_attributes"] == []

    info_conf = get_topo_inputs_config
    info_conf["terrain_attributes"] = ["hillshade", "slope", "curvature"]
    doc = schemas.validate_configuration(info_conf, schemas.TOPO_SCHEMA)
    assert doc["terrain_attributes"] == ["hillshade", "slope", "curvature"]



# Copyright (c) 2025 Centre National d'Etudes Spatiales (CNES).
#
# This file is part of the xDEM project:
# https://github.com/glaciohack/xdem
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Schema constants and validation function
"""
import logging
import os
from typing import Any, Dict
from urllib.error import HTTPError, URLError

from cerberus import Validator

from xdem.vcrs import _vcrs_from_user_input


class CustomValidator(Validator):  # type: ignore
    def _validate_path_exists(self, path_exists: bool, field: str, value: str) -> bool:
        """
        {'type': 'boolean'}
        """
        if value is not None:
            if path_exists and not os.path.exists(value):
                self._error(field, f"Path does not exist: {value}")
        return True

    def _validate_crs(self, crs: bool, field: str, value: str | int) -> bool:
        """
        {'type': 'boolean'}
        """
        if value is not None:
            try:
                _vcrs_from_user_input(value)
                return True
            except (ValueError, TypeError, ConnectionResetError, HTTPError, URLError):
                logging.error(
                    f"'{field}' field is not valid. See: https://xdem.readthedocs.io/en/stable/vertical_ref.html"
                )
                return False
        return True


INPUTS_DEM = {
    "path_to_elev": {"type": "string", "required": True, "path_exists": True},
    "force_source_nodata": {"type": ["integer", "float"], "required": False, "nullable": True},
    "path_to_mask": {"type": "string", "required": False, "path_exists": True, "nullable": True},
    "from_vcrs": {"type": ["integer", "string"], "required": False, "nullable": True, "crs": True, "default": None},
    "to_vcrs": {"type": ["integer", "string"], "required": False, "nullable": True, "crs": True, "default": None},
}

COREG_METHODS = ["NuthKaab", "DhMinimize", "VerticalShift", "DirectionalBias", "TerrainBias", "LZD", None]

STATS_METHODS = [
    "mean",
    "median",
    "max",
    "min",
    "sum",
    "sumofsquares",
    "90thpercentile",
    "le90",
    "nmad",
    "rmse",
    "std",
    "standarddeviation",
    "validcount",
    "totalcount",
    "percentagevalidpoints",
]

TERRAIN_ATTRIBUTES_DEFAULT = ["slope", "aspect", "curvature"]

TERRAIN_ATTRIBUTES = [
    "slope",
    "aspect",
    "hillshade",
    "curvature",
    "planform_curvature",
    "profile_curvature",
    "maximum_curvature",
    "terrain_ruggedness_index",
    "topographic_position_index",
    "roughness",
    "rugosity",
    "fractal_roughness",
]


def make_coreg_step(required: bool = False, default_method: str = None) -> Validator.schema:
    """
    Create a coreg schema step to avoid repetition for a step
    :param required: is a step required
    :param default_method: coreg default method
    """
    step_schema = {
        "type": "dict",
        "required": required,
        "schema": {
            "method": {
                "type": "string",
                "allowed": COREG_METHODS,
                "required": True if required else False,
                "nullable": False if required else True,
            },
            "extra_information": {"type": "dict", "required": False, "nullable": True},
        },
    }
    if default_method:
        step_schema["default"] = {"method": default_method}
    return step_schema


def validate_configuration(user_config: dict[str, Any], schema: Dict[str, Any]) -> Dict[str, Any]:
    """
    Validate the configuration:
    :param user_config: Configuration dict or YAML string
    :param schema: Schema dict for validating configuration
    :return: Completed configuration dictionary
    """
    validator = CustomValidator(schema)
    if not validator.validate(user_config):
        for field, errors in validator.errors.items():
            raise ValueError(f"User configuration mistakes in '{field}': {errors}")

    if "statistics" not in validator.document:
        validator.document["statistics"] = STATS_METHODS

    if "terrain_attributes" not in validator.document and "coregistration" not in validator.document:
        validator.document["terrain_attributes"] = TERRAIN_ATTRIBUTES_DEFAULT

    return validator.document


ACCURACY_SCHEMA = {
    "inputs": {
        "type": "dict",
        "required": True,
        "schema": {
            "reference_elev": {"type": "dict", "schema": INPUTS_DEM, "required": False, "nullable": True},
            "to_be_aligned_elev": {"type": "dict", "schema": INPUTS_DEM, "required": True},
        },
    },
    "outputs": {
        "type": "dict",
        "required": False,
        "default": {"path": "outputs", "level": 1, "output_grid": "reference_elev"},
        "schema": {
            "path": {"type": "string", "required": True, "default": "outputs"},
            "level": {"type": "integer", "default": 1, "required": False, "allowed": [1, 2]},
            "output_grid": {
                "type": "string",
                "allowed": ["reference_elev", "to_be_aligned_elev"],
                "default": "reference_elev",
                "required": False,
            },
        },
    },
    "coregistration": {
        "type": "dict",
        "required": False,
        "default": {"step_one": {"method": "NuthKaab"}, "sampling_grid": "reference_elev"},
        "schema": {
            "step_one": make_coreg_step(default_method="NuthKaab"),
            "step_two": make_coreg_step(required=False),
            "step_three": make_coreg_step(required=False),
            "sampling_grid": {
                "type": "string",
                "allowed": ["reference_elev", "to_be_aligned_elev"],
                "default": "reference_elev",
                "required": False,
            },
            "process": {"type": "boolean", "default": True, "required": False},
        },
    },
    "statistics": {"type": "list", "required": False, "allowed": STATS_METHODS, "nullable": True},
}

TOPO_SCHEMA = {
    "inputs": {
        "type": "dict",
        "required": True,
        "schema": {
            "reference_elev": {"type": "dict", "schema": INPUTS_DEM, "required": False},
        },
    },
    "statistics": {"type": "list", "required": False, "allowed": STATS_METHODS, "nullable": True},
    "terrain_attributes": {
        "required": False,
        "default": TERRAIN_ATTRIBUTES_DEFAULT,
        "nullable": True,
        "anyof": [
            {
                "type": "list",
                "schema": {
                    "type": "string",
                    "allowed": TERRAIN_ATTRIBUTES,
                },
            },
            {
                "type": "dict",
                "keysrules": {"type": "string", "allowed": TERRAIN_ATTRIBUTES},
                "valuesrules": {
                    "type": "dict",
                    "schema": {
                        "extra_information": {"type": "dict", "required": False},
                    },
                },
            },
        ],
    },
    "outputs": {
        "type": "dict",
        "default": {"path": "outputs", "level": 1},
        "schema": {
            "path": {"type": "string", "default": "outputs"},
            "level": {"type": "integer", "default": 1, "required": False, "allowed": [1, 2]},
        },
    },
}

COMPLETE_CONFIG_ACCURACY = {
    "inputs": {
        "reference_elev": {
            "path_to_elev": "",
            "force_source_nodata": None,
            "from_vcrs": None,
            "to_vcrs": None,
        },
        "to_be_aligned_elev": {
            "path_to_elev": "",
            "force_source_nodata": None,
            "from_vcrs": None,
            "to_vcrs": None,
            "path_to_mask": None,
        },
    },
    "outputs": {
        "level": 1,
        "path": "outputs",
        "output_grid": "reference_elev",
    },
    "coregistration": {
        "sampling_grid": "reference_elev",
        "step_one": {
            "method": "NuthKaab",
            "extra_information": None,
        },
        "step_two": {
            "method": None,
            "extra_information": None,
        },
        "step_three": {
            "method": None,
            "extra_information": None,
        },
    },
    "statistics": [
        "mean",
        "median",
        "max",
        "min",
        "sum",
        "sumofsquares",
        "90thpercentile",
        "le90",
        "nmad",
        "rmse",
        "std",
        "standarddeviation",
        "validcount",
        "totalcount",
        "percentagevalidpoints",
    ],
}

COMPLETE_CONFIG_TOPO = {
    "inputs": {
        "reference_elev": {
            "path_to_elev": "",
            "force_source_nodata": None,
            "from_vcrs": None,
            "to_vcrs": None,
            "path_to_mask": None,
        },
    },
    "outputs": {"level": 1, "path": "outputs"},
    "statistics": [
        "mean",
        "median",
        "max",
        "min",
        "sum",
        "sumofsquares",
        "90thpercentile",
        "le90",
        "nmad",
        "rmse",
        "std",
        "standarddeviation",
        "validcount",
        "totalcount",
        "percentagevalidpoints",
    ],
    "terrain_attributes": ["slope", "aspect", "curvature"],
}



# Copyright (c) 2025 Centre National d'Etudes Spatiales (CNES).
#
# This file is part of the xDEM project:
# https://github.com/glaciohack/xdem
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test Topo class
"""
# mypy: disable-error-code=no-untyped-def
from pathlib import Path

import pytest
from rasterio import Affine
from rasterio.coords import BoundingBox

import xdem
from xdem.workflows import Topo
from xdem.workflows.schemas import STATS_METHODS
from xdem.workflows.workflows import Workflows

pytestmark = pytest.mark.filterwarnings("ignore::UserWarning")


def test_init_topo_summary(get_topo_inputs_config, tmp_path, list_default_terrain_attributes):
    """
    Test Topo class initialization
    """
    user_config = get_topo_inputs_config
    user_config["outputs"] = {"path": str(tmp_path)}
    workflows = Topo(user_config)

    assert isinstance(workflows, Workflows)
    assert isinstance(workflows, Topo)
    assert Path(tmp_path / "plots").joinpath("elev_map.png").exists()
    assert Path(tmp_path / "plots").joinpath("masked_elev_map.png").exists()
    assert workflows.config_attributes == list_default_terrain_attributes

    user_config = get_topo_inputs_config
    user_config["outputs"] = {"path": str(tmp_path)}
    user_config["terrain_attributes"] = []
    workflows = Topo(user_config)
    assert workflows.config_attributes == []

    user_config = get_topo_inputs_config
    user_config["outputs"] = {"path": str(tmp_path)}
    user_config["terrain_attributes"] = {
        "hillshade": {"extra_information": {"method": "ZevenbergThorne", "azimuth": 90}}
    }
    workflows = Topo(user_config)
    assert workflows.config_attributes == {
        "hillshade": {"extra_information": {"method": "ZevenbergThorne", "azimuth": 90}}
    }
    assert workflows.list_attributes == ["hillshade"]


def test_generate_terrain_attributes(tmp_path, get_topo_inputs_config, list_default_terrain_attributes):
    """
    Test generate_terrain_attributes function
    """
    user_config = get_topo_inputs_config
    user_config["outputs"] = {"path": str(tmp_path)}
    workflows = Topo(user_config)

    workflows.generate_terrain_attributes_tiff()

    for attr in list_default_terrain_attributes:
        assert Path(tmp_path / "rasters").joinpath(f"{attr}.tif").exists()


def test_generate_terrain_attributes_level_2(tmp_path, get_topo_inputs_config, list_default_terrain_attributes):
    """
    Test generate_terrain_attributes function
    """
    user_config = get_topo_inputs_config
    user_config["outputs"] = {"path": str(tmp_path), "level": 2}
    workflows = Topo(user_config)
    workflows.run()

    for attr in list_default_terrain_attributes:
        assert Path(tmp_path / "rasters").joinpath(f"{attr}.tif").exists()


def test_generate_no_terrain_attributes(tmp_path, get_topo_inputs_config, list_default_terrain_attributes):
    """
    Test generate_terrain_attributes function
    """
    user_config = get_topo_inputs_config
    user_config["outputs"] = {"path": str(tmp_path), "level": 2}
    user_config["terrain_attributes"] = None
    workflows = Topo(user_config)
    workflows.run()

    for attr in list_default_terrain_attributes:
        assert not Path(tmp_path / "rasters").joinpath(f"{attr}.tif").exists()


def test_generate_terrain_attributes_png(tmp_path, get_topo_inputs_config):
    """
    Test generate_terrain_attributes_png function
    """
    user_config = get_topo_inputs_config
    user_config["outputs"] = {"path": str(tmp_path)}
    workflows = Topo(user_config)
    workflows.run()

    workflows.generate_terrain_attributes_png()
    assert Path(tmp_path / "plots").joinpath("terrain_attributes_map.png").exists()


def test_run(get_topo_inputs_config, tmp_path):
    """
    Test run function
    """

    user_config = get_topo_inputs_config
    user_config["outputs"] = {"path": str(tmp_path)}
    workflows = Topo(user_config)
    workflows.run()

    assert Path(tmp_path / "tables").joinpath("stats_elev_stats.csv").exists()
    assert Path(tmp_path / "tables").joinpath("stats_elev_mask_stats.csv").exists()
    assert Path(tmp_path).joinpath("report.html").exists()
    # Check subdictionaries content, except exact stats values in case test data/algorithms slightly changes,
    # and as those are already tested separately
    # 1/ Input information
    assert workflows.dico_to_show[0] == (
        "Information about inputs",
        {
            "reference_elev": {
                "path_to_elev": xdem.examples.get_path("longyearbyen_tba_dem"),
                "from_vcrs": None,
                "path_to_mask": xdem.examples.get_path("longyearbyen_glacier_outlines"),
                "to_vcrs": None,
            }
        },
    )
    # 2/ Elevation information
    assert workflows.dico_to_show[1] == (
        "Elevation information",
        {
            "Data types": "float32",
            "Driver": "GTiff",
            "Filename": xdem.examples.get_path("longyearbyen_tba_dem"),
            "Grid size": None,
            "Height": 985,
            "Nodata Value": -9999.0,
            "Number of band": (1,),
            "Pixel interpretation": "Area",
            "Pixel size": (20.0, 20.0),
            "Transform": Affine(20.0, 0.0, 502810.0, 0.0, -20.0, 8674030.0),
            "Width": 1332,
            "Bounds": BoundingBox(left=502810.0, bottom=8654330.0, right=529450.0, top=8674030.0),
        },
    )
    # 3/ Statistics names
    assert workflows.dico_to_show[2][0] == "Global statistics"
    assert list(workflows.dico_to_show[2][1].keys()) == STATS_METHODS
    assert workflows.dico_to_show[3][0] == "Mask statistics"
    assert list(workflows.dico_to_show[3][1].keys()) == STATS_METHODS



# Copyright (c) 2025 Centre National d'Etudes Spatiales (CNES).
#
# This file is part of the xDEM project:
# https://github.com/glaciohack/xdem
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test DiffAnalysis class
"""
# mypy: disable-error-code=no-untyped-def
from pathlib import Path

import geoutils as gu
import pytest

import xdem
from xdem.workflows import Accuracy
from xdem.workflows.workflows import Workflows

pytestmark = pytest.mark.filterwarnings("ignore::UserWarning")


def test_init_diff_analysis(get_accuracy_object_with_run, tmp_path):
    """
    Test initialization of accuracy class
    """
    workflows = get_accuracy_object_with_run

    assert isinstance(workflows, Workflows)
    assert isinstance(workflows, Accuracy)
    assert Path(tmp_path / "plots").joinpath("reference_elev_map.png").exists()
    assert Path(tmp_path / "plots").joinpath("to_be_aligned_elev_map.png").exists()
    assert Path(tmp_path / "plots").joinpath("reference_elev_map.png").exists()
    dem = xdem.DEM(xdem.examples.get_path("longyearbyen_tba_dem"))
    mask = gu.Vector(xdem.examples.get_path("longyearbyen_glacier_outlines"))
    inlier_mask = ~mask.create_mask(dem)
    assert workflows.inlier_mask == inlier_mask


def test__get_reference_elevation(get_accuracy_inputs_config, tmp_path):
    """
    Test _get_reference_elevation function
    """

    user_config = get_accuracy_inputs_config
    user_config["outputs"] = {"path": str(tmp_path)}
    workflows = Accuracy(user_config)

    with pytest.raises(NotImplementedError, match="For now it doesn't working, please add a reference DEM"):
        workflows._get_reference_elevation()

    user_config = get_accuracy_inputs_config
    user_config["outputs"] = {"path": str(tmp_path)}
    user_config["inputs"]["reference_elev"] = None

    with pytest.raises(NotImplementedError, match="For now it doesn't working, please add a reference DEM"):
        _ = Accuracy(user_config)


@pytest.mark.skip("Not implemented")
def test__compute_coregistration():
    """
    Test _compute_coregistration function
    """


def test__compute_reproj(get_accuracy_inputs_config, tmp_path):
    """
    Test _compute_reproj function
    """
    user_config = get_accuracy_inputs_config
    user_config["outputs"] = {"path": str(tmp_path)}
    user_config["coregistration"] = {"process": False}
    user_config["coregistration"]["sampling_grid"] = "to_be_aligned_elev"
    workflows = Accuracy(user_config)
    workflows.run()
    src, target = workflows.reference_elev, workflows.to_be_aligned_elev
    gt_reprojected = src.reproject(target, silent=True)

    assert workflows.reference_elev == gt_reprojected


def test__get_stats(get_accuracy_inputs_config, tmp_path):
    """
    Test _get_stats function
    """
    user_config = get_accuracy_inputs_config
    user_config["outputs"] = {"path": str(tmp_path)}
    workflows = Accuracy(user_config)

    dem = xdem.DEM(xdem.examples.get_path("longyearbyen_tba_dem"))
    stats_gt = dem.get_stats(
        [
            "mean",
            "median",
            "max",
            "min",
            "sum",
            "sumofsquares",
            "90thpercentile",
            "le90",
            "nmad",
            "rmse",
            "std",
            "standarddeviation",
            "validcount",
            "totalcount",
            "percentagevalidpoints",
        ]
    )

    # Aliases for nicer CSV headers
    aliases = {
        "mean": "Mean",
        "median": "Median",
        "max": "Maximum",
        "min": "Minimum",
        "sum": "Sum",
        "sumofsquares": "Sum of squares",
        "90thpercentile": "90th percentile",
        "le90": "LE90",
        "nmad": "NMAD",
        "rmse": "RMSE",
        "std": "STD",
        "standarddeviation": "Standard deviation",
        "validcount": "Valid count",
        "totalcount": "Total count",
        "percentagevalidpoints": "Percentage valid points",
    }

    stats_gt = {aliases.get(k, k): v for k, v in stats_gt.items()}
    assert workflows._get_stats(dem) == stats_gt


def test__compute_histogram(get_accuracy_object_with_run, tmp_path):
    """
    Test _compute_histogram function
    """

    _ = get_accuracy_object_with_run

    assert Path(tmp_path / "plots").joinpath("elev_diff_histo.png").exists()


@pytest.mark.parametrize(
    "level",
    [1, 2],
)
def test_run(get_accuracy_inputs_config, tmp_path, level):
    """
    Test run function
    """

    user_config = get_accuracy_inputs_config
    user_config["outputs"] = {"path": str(tmp_path), "level": level}
    workflows = Accuracy(user_config)
    workflows.run()

    assert Path(tmp_path / "tables").joinpath("aligned_elev_stats.csv").exists()

    assert Path(tmp_path / "plots").joinpath("diff_elev_after_coreg.png").exists()
    assert Path(tmp_path / "plots").joinpath("diff_elev_before_coreg.png").exists()
    assert Path(tmp_path / "plots").joinpath("elev_diff_histo.png").exists()
    assert Path(tmp_path / "plots").joinpath("masked_elev_map.png").exists()
    assert Path(tmp_path / "plots").joinpath("reference_elev_map.png").exists()
    assert Path(tmp_path / "plots").joinpath("to_be_aligned_elev_map.png").exists()

    assert Path(tmp_path / "rasters").joinpath("aligned_elev.tif").exists()

    assert Path(tmp_path).joinpath("report.html").exists()
    # sometimes the PDF creation fails for no reason
    # assert Path(tmp_path).joinpath("report.pdf").exists()
    assert Path(tmp_path).joinpath("used_config.yaml").exists()

    csv_files_level_1 = [
        "diff_elev_after_coreg_stats.csv",
        "diff_elev_before_coreg_stats.csv",
    ]

    csv_files_level_2 = [
        "reference_elev_stats.csv",
        "to_be_aligned_elev_stats.csv",
    ]

    raster_files = ["diff_elev_after_coreg.tif", "diff_elev_before_coreg.tif", "to_be_aligned_elev_reprojected.tif"]

    if level == 1:
        for file in csv_files_level_1:
            assert (Path(tmp_path) / "tables" / file).exists()
        for file in csv_files_level_2:
            assert not (Path(tmp_path) / "tables" / file).exists()
        for file in raster_files:
            assert not (Path(tmp_path) / "rasters" / file).exists()

    if level == 2:
        for file in csv_files_level_1:
            assert (Path(tmp_path) / "tables" / file).exists()
        for file in csv_files_level_2:
            assert (Path(tmp_path) / "tables" / file).exists()
        for file in raster_files:
            assert (Path(tmp_path) / "rasters" / file).exists()


@pytest.mark.parametrize(
    "level",
    [1, 2],
)
def test_run_without_coreg(get_accuracy_inputs_config, tmp_path, level):
    """
    Test run function
    """

    user_config = get_accuracy_inputs_config
    user_config["outputs"] = {"path": str(tmp_path), "level": level}
    user_config["coregistration"] = {"process": False}
    workflows = Accuracy(user_config)
    workflows.run()

    assert Path(tmp_path / "tables").joinpath("diff_elev_stats.csv").exists()

    assert Path(tmp_path / "plots").joinpath("diff_elev.png").exists()
    assert not Path(tmp_path / "plots").joinpath("diff_elev_before_coreg.png").exists()
    assert not Path(tmp_path / "plots").joinpath("elev_diff_histo.png").exists()
    assert Path(tmp_path / "plots").joinpath("masked_elev_map.png").exists()
    assert Path(tmp_path / "plots").joinpath("reference_elev_map.png").exists()
    assert Path(tmp_path / "plots").joinpath("to_be_aligned_elev_map.png").exists()

    assert not Path(tmp_path / "rasters").joinpath("aligned_elev.tif").exists()

    assert Path(tmp_path).joinpath("report.html").exists()
    # sometimes the PDF creation fails for no reason
    # assert Path(tmp_path).joinpath("report.pdf").exists()
    assert Path(tmp_path).joinpath("used_config.yaml").exists()

    csv_files = [
        "diff_elev_stats.csv",
        "reference_elev_stats.csv",
        "to_be_aligned_elev_stats.csv",
    ]

    raster_files = ["diff_elev.tif"]

    if level == 1:
        for file in csv_files:
            assert (Path(tmp_path) / "tables" / file).exists()
        for file in raster_files:
            assert not (Path(tmp_path) / "rasters" / file).exists()

    if level == 2:
        for file in csv_files:
            assert (Path(tmp_path) / "tables" / file).exists()
        for file in raster_files:
            assert not (Path(tmp_path) / "rasters" / file).exists()


def test_create_html(tmp_path, get_accuracy_object_with_run):
    """
    Test create_html function
    """
    _ = get_accuracy_object_with_run

    assert Path(tmp_path).joinpath("report.html").exists()


def test_mask_init(tmp_path, get_accuracy_inputs_config):
    """
    Test mask initialization
    """
    user_config = get_accuracy_inputs_config
    user_config["outputs"] = {"path": str(tmp_path)}
    del user_config["inputs"]["reference_elev"]["path_to_mask"]
    workflows = Accuracy(user_config)
    dem = xdem.DEM(xdem.examples.get_path("longyearbyen_tba_dem"))
    mask = gu.Vector(xdem.examples.get_path("longyearbyen_glacier_outlines"))
    inlier_mask = ~mask.create_mask(dem)
    assert workflows.inlier_mask == inlier_mask
    assert Path(tmp_path / "plots").joinpath("masked_elev_map.png").exists()
